@@ -7,7 +7,9 @@
 
 #include <config.h>
 
-#include "lib/autotrust.h"
+#include "src/autotrust.h"
+#include "src/options.h"
+#include "src/util.h"
 
 /* Our environment in which we can perform automated updates of trust anchors */
 autotrust_t* env = NULL;
@@ -51,6 +53,8 @@ initialize_environment(void)
 		return NULL;
 	/* nothing changed (yet) */
 	new_env->changed = 0;
+	/* no signal received (yet) */
+	new_env->sig_quit = 0;
 	/* default log to stderr, verbosity 1 */
 	open_logfile(NULL, 1);
 	/* done */
@@ -61,9 +65,6 @@ initialize_environment(void)
 static void
 usage(void)
 {
-	const char* version = "0.2";
-	const char* bugreport = "matthijs@nlnetlabs.nl";
-
 	fprintf(stderr, "Usage: %s [OPTIONS]\n", AUTOTRUST_NAME);
 	fprintf(stderr, "Automated updates of trust anchors (RFC 5011).\n\n");
 	fprintf(stderr,
@@ -78,7 +79,265 @@ usage(void)
 	);
 
 	fprintf(stderr, "Version %s, Report bugs to <%s>.\n",
-		version, bugreport);
+		PACKAGE_VERSION, PACKAGE_BUGREPORT);
+}
+
+/** Write pidfile */
+static int
+writepid(char* pidfile, pid_t pid)
+{
+	FILE * fd;
+	char pidbuf[32];
+	size_t result = 0, size = 0;
+
+	snprintf(pidbuf, sizeof(pidbuf), "%lu\n", (unsigned long) pid);
+
+	if ((fd = fopen(pidfile, "w")) ==  NULL )
+	{
+		error("cannot open pidfile %s: %s", pidfile, strerror(errno));
+		return -1;
+	}
+
+	size = strlen(pidbuf);
+	if (size == 0)
+		result = 1;
+	result = fwrite((const void*) pidbuf, 1, size, fd);
+	if (result == 0)
+	{
+		error("write to pidfile failed: %s", strerror(errno));
+		result = 0;
+    }
+	else if (result < size)
+	{
+		error("short write to pidfile (disk full?)");
+        result = 0;
+	}
+	else
+		result = 1;
+
+	if (!result)
+	{
+		error("cannot write pidfile %s: %s", pidfile, strerror(errno));
+		fclose(fd);
+		return -1;
+	}
+
+	fclose(fd);
+	/* chown? */
+
+    return 0;
+}
+
+/** Signal handling. */
+void
+sig_handler(int sig)
+{
+	switch (sig)
+	{
+		case SIGTERM:
+			debug(("sigterm received"));
+			env->sig_quit = 1;
+			break;
+		case SIGHUP:
+			debug(("sighup received"));
+			env->sig_quit = 1;
+			break;
+		default:
+			break;
+	}
+	return;
+}
+
+/** Daemonize setup. */
+static pid_t
+setup_daemon(void)
+{
+	pid_t pid = -1;
+	struct sigaction action;
+
+	switch ((pid = fork()))
+	{
+		case 0: /* child */
+			break;
+		case -1: /* error */
+			error("fork() failed: %s", strerror(errno));
+			free_environment(1);
+		default: /* parent is done */
+			debug(("fork() ok, parent quit"));
+			free_environment(0);
+	}
+	if (setsid() == -1)
+	{
+		error("setsid() failed: %s", strerror(errno));
+		free_environment(1);
+	}
+
+	/* setup signal handing */
+	action.sa_handler = sig_handler;
+	sigfillset(&action.sa_mask);
+	action.sa_flags = 0;
+	sigaction(SIGHUP, &action, NULL);
+
+	pid = getpid();
+	if (writepid(env->options->pidfile, pid) == -1)
+		free_environment(1);
+
+	return pid;
+}
+
+/** Poll the authoritative servers for DNSKEYs */
+static int
+active_refresh(tp_t* tp)
+{
+	struct ub_ctx* ctx = NULL;
+	int return_status = STATUS_OK;
+
+	verbos(2, "active refresh for zone %s", tp->name);
+	ctx = ub_ctx_create();
+	if (!ctx)
+	{
+		error("failed to create unbound context, aborting "
+		      "active refresh for %s", tp->name);
+		return STATUS_ERR_QUERY;
+	}
+
+	/* configuration settings */
+	return_status = set_resolver_options(env->options, tp, ctx);
+	if (return_status != STATUS_OK)
+		warning("could not set libunbound resolver options: %s, continuing "
+				"with default settings", status_by_id(return_status));
+	/* resolve and auto update */
+	return_status = query_dnskeys(ctx, tp);
+	if (return_status != STATUS_OK)
+		error("active refresh for zone %s failed: %s", tp->name,
+			status_by_id(return_status));
+	else /* perform RFC 5011 statetable */
+		return_status = do_statetable(tp, env->options);
+
+	if (return_status == STATUS_CHANGED)
+	{
+		env->changed = 1;
+		return_status = STATUS_OK;
+	}
+	if (return_status != STATUS_OK)
+		error("failed to update state for %s: %s", tp->name,
+			status_by_id(return_status));
+	/* delete context */
+	ub_ctx_delete(ctx);
+
+	return return_status;
+}
+
+/** Update the files on disk. */
+static int
+update_files(void)
+{
+	int return_status = STATUS_OK;
+	int return_status2 = STATUS_OK;
+
+	debug(("update files..."));
+	return_status = update_state(env->trustpoints, env->options);
+	if (return_status != STATUS_OK)
+		error("could not update state file: %s", status_by_id(return_status));
+	if (return_status == STATUS_OK && env->changed)
+	{
+		return_status = update_trustanchor_files(env->trustpoints,
+			env->options);
+		if (return_status != STATUS_OK)
+			error("could not update (all) trust anchor files: %s",
+						status_by_id(return_status));
+		else
+		{
+			return_status2 = update_trustedkeys_files(env->trustpoints,
+				env->options->trustedkeysfiles);
+			if (return_status2 != STATUS_OK)
+			{
+				error("could not update (all) trusted keys files");
+				return_status = return_status2;
+			}
+		}
+	}
+	return return_status;
+}
+
+/** Do a single run over all trust points. */
+static int
+do_single_run(time_t* wait)
+{
+	int return_status = STATUS_OK;
+	tp_t* tp = NULL;
+	rbnode_t* node = NULL;
+	time_t now = time(NULL), refresh = 0;
+	strlist_t* list = NULL;
+
+	/* setting: state-file: load state from this file */
+	return_status = load_state(env->trustpoints, env->options->statefile);
+	if (return_status != STATUS_OK)
+		error("could not load state: %s", status_by_id(return_status));
+	debug(("trust anchors loaded"));
+
+#ifdef TDEBUG
+	debug_print_trustpoints(env->trustpoints);
+#endif /* TDEBUG */
+
+	/* resolve and update */
+	debug(("start updating"));
+	if (env->trustpoints)
+		node = rbt_first(env->trustpoints->sep);
+	while (node != NULL)
+	{
+		/* trust point */
+		tp = (tp_t*) node->key;
+		if (!tp)
+		{
+			node = rbt_successor(node);
+			continue;
+		}
+		/* calculate first next refresh */
+		refresh = next_refresh(tp);
+		if (now < refresh)
+		{
+			*wait = min(*wait, refresh);
+			node = rbt_successor(node);
+			continue;
+		}
+
+		/* query for new DNSKEYs */
+		return_status = active_refresh(tp);
+
+		/* update wait period */
+		refresh = next_refresh(tp);
+		if (now < refresh)
+			*wait = min(*wait, refresh);
+
+		/* next zone */
+		node = rbt_successor(node);
+	}
+
+	/* update trust anchor files */
+	return_status = update_files();
+
+	/* setting: resolver-pidfile: signal resolver if trust
+	 * anchors were updated.
+	 */
+	if (return_status == STATUS_OK && env->options)
+	{
+		list = env->options->resolvers;
+		while (list)
+		{
+			signal_resolver(list->str);
+			list = list->next;
+		}
+
+		list = env->options->reloadcmd;
+		while (list)
+		{
+			reload_resolver(list->str);
+			list = list->next;
+		}
+    }
+
+	return return_status;
 }
 
 /*
@@ -107,12 +366,11 @@ usage(void)
 int
 main(int argc, char* argv[])
 {
-	/* read commandline arguments */
-	int c, r;
-	/* everything is ok until so far */
+	/* some necessary variables */
+	int c, r, running;
 	int return_status = STATUS_OK;
-	/* walk the trust point set */
-	rbnode_t* node = NULL;
+	time_t wait = 0, now = time(NULL);
+	pid_t pid;
 
 	/* initialize the autotrust environment */
 	env = initialize_environment();
@@ -121,23 +379,17 @@ main(int argc, char* argv[])
 	debug(("autotrust environment initialized"));
 
 	/* read command line options */
-#if 0
 	while ((c = getopt(argc, argv, "c:dhv:")) != -1)
-#else
-	while ((c = getopt(argc, argv, "c:hv:")) != -1)
-#endif
 	{
 		switch (c)
 		{
 			case 'c': /* config file */
 				env->configfile = optarg;
 				break;
-#if 0
 			case 'd': /* daemonize */
 				r = set_cfg_option_int(env->options,
 								"daemonize", 1);
 				break;
-#endif
 			case 'v': /* verbosity */
 				r = set_cfg_option_int(env->options,
 						"verbosity", atoi(optarg));
@@ -168,44 +420,51 @@ main(int argc, char* argv[])
 	debug_print_options(env->options);
 #endif /* TDEBUG */
 
-	/* setting: working-dir: change to the configured working directory */
+	/* apply options */
+	if (env->options)
+	{
+		if (env->options->use_syslog)
+		{
+#ifdef HAVE_SYSLOG_H
+			openlog("autotrust", LOG_NDELAY, LOG_DAEMON);
+			set_logfunction(log_syslog);
+#else
+			warning("syslog not supported, logging to stderr");
+#endif /* HAVE_SYSLOG_H */
+		}
+		else
+			open_logfile(env->options->logfile, env->options->verbosity);
+
+		debug(("logfile %s opened with verbosity %i",
+			env->options->use_syslog?"syslog":(env->options->logfile?env->options->logfile:"stderr"),
+			env->options->verbosity));
+	}
+
 	if (env->options && env->options->workingdir &&
 		env->options->workingdir[0])
 	{
 		if (chdir(env->options->workingdir) != 0)
 		{
-			verbos(3, "cannot chdir to %s: %s",
-				env->options->workingdir, strerror(errno));
+			verbos(3, "cannot chdir to %s: %s", env->options->workingdir,
+				strerror(errno));
 			free_environment(STATUS_ERR_CHDIR);
 		}
-		debug(("working directory changed to %s",
-			env->options->workingdir));
+		debug(("working directory changed to %s", env->options->workingdir));
 	}
 
-	/* setting: log-file: log messages to file */
-	/* setting: verbosity: verbosity level */
-	if (env->options)
-	{
-		open_logfile(env->options->logfile, env->options->verbosity);
-		debug(("logfile %s opened",
-			env->options->logfile?env->options->logfile:"stderr"));
-	}
-
-	/* [TODO] daemonize */
-
-	/* setting: trust-anchor-file: load trust anchor files */
+	/* trust anchor files */
 	return_status = load_trustanchor_files(env->trustpoints,
 		env->options->trustanchorfiles);
 	if (return_status != STATUS_OK)
 		error("could not load trust anchor files: %s",
 			status_by_id(return_status));
-	/* setting: trusted-keys-file: load trusted keys files */
+
 	return_status = load_trustedkeys_files(env->trustpoints,
 		env->options->trustedkeysfiles);
 	if (return_status != STATUS_OK)
 		error("could not load trusted keys: %s",
 			status_by_id(return_status));
-	/* setting: trust-anchor: load additional trust anchors */
+
 #if 0
 	return_status = load_trustanchor_options(env->trustpoints,
 		env->options->trustanchors);
@@ -213,108 +472,45 @@ main(int argc, char* argv[])
 		error("could not load trust anchor options: %s",
 			status_by_id(return_status));
 #endif
-	/* setting: state-file: load state from this file */
-	return_status = load_state(env->trustpoints, env->options->statefile);
-	if (return_status != STATUS_OK)
-		error("could not load state: %s", status_by_id(return_status));
-	debug(("trust anchors loaded"));
 
-#ifdef TDEBUG
-	debug_print_trustpoints(env->trustpoints);
-#endif /* TDEBUG */
+	/* if we run as daemon, some extra work is needed */
+	running = env->options->daemonize;
+	if (running)
+		pid = setup_daemon();
 
-	/* resolve and update */
-	debug(("start updating"));
-	if (env->trustpoints)
-		node = rbt_first(env->trustpoints->sep);
-	while (node != NULL)
+	/* do a single or multiple runs */
+	do
 	{
-		/* libunbound context */
-		struct ub_ctx* ctx;
-		/* trust point */
-		tp_t* tp = (tp_t*) node->key;
-		if (!tp)
-			continue;
+		/* if a sigterm is received, the party is over */
+		if (env->sig_quit) { running = 0; break; }
 
-		/* query for new DNSKEYs */
-		verbos(2, "active refresh for zone %s", tp->name);
-		ctx = ub_ctx_create();
-		if (!ctx)
+		/* have we waited long enough? */
+		now = time(NULL);
+		if (now < wait && !env->sig_quit)
 		{
-			error("failed to create unbound context, aborting "
-			      "active refresh for %s", tp->name);
+			sleep(wait-now);
 			continue;
 		}
-		/* configuration settings */
-		return_status = set_resolver_options(env->options, tp, ctx);
-		if (return_status != STATUS_OK)
-			warning("could not set libunbound resolver options: "
-				"%s, continuing with default settings",
-						status_by_id(return_status));
 
-		/* resolve and auto update */
-		return_status = query_dnskeys(ctx, tp);
-		if (return_status != STATUS_OK)
-			error("active refresh for zone %s failed: %s", tp->name,
-						status_by_id(return_status));
-
-		/* perform RFC 5011 statetable */
-		return_status = do_statetable(tp, env->options);
-		if (return_status == STATUS_CHANGED)
+		/* now we have, set a new wait period */
+		wait = now + FIFTEEN_DAYS;
+		/* do an active refresh for all trust points */
+		return_status = do_single_run(&wait);
+		if (running)
 		{
-			env->changed = 1;
-			return_status = STATUS_OK;
+			debug(("waiting %i seconds to next refresh", (wait-now)));
+			flush();
 		}
-		if (return_status != STATUS_OK)
-			error("failed to update state for %s: %s", tp->name,
-						status_by_id(return_status));
-		/* delete context */
-		ub_ctx_delete(ctx);
-		/* next zone */
-		node = rbt_successor(node);
-	}
-
-	/* update trunst anchor files */
-	debug(("update files..."));
-	return_status = update_state(env->trustpoints, env->options);
-	if (return_status != STATUS_OK)
-		error("could not update state file: %s",
-						status_by_id(return_status));
-
-	if (return_status == STATUS_OK && env->changed)
-	{
-		return_status = update_trustanchor_files(env->trustpoints,
-			env->options);
-		if (return_status != STATUS_OK)
-			error("could not update (all) trust anchor files: %s",
-						status_by_id(return_status));
-		else {
-			int return_status2 =
-				update_trustedkeys_files(env->trustpoints,
-				env->options->trustedkeysfiles);
-			if (return_status2 != STATUS_OK) {
-				error("could not update (all) trusted keys "
-					"files");
-				return_status = return_status2;
-			}
-		}
-
-		/* setting: resolver-pidfile: signal resolver if trust
-		 * anchors were updated.
-		 */
-		if (return_status == STATUS_OK && env->options)
-		{
-			strlist_t* list = env->options->resolverpidfiles;
-			while (list)
-			{
-				signal_resolver(list->str);
-	                        list = list->next;
-			}
-        	}
 	}
+	while (running);
 
 	/* done */
 	debug(("done"));
+#ifdef HAVE_SYSLOG_H
+	if (env->options->use_syslog)
+		closelog();
+#endif /* HAVE_SYSLOG_H */
+
 	free_environment(return_status);
 
 	/* not reached */
